import pytest
from hana.core import FileSet, File

def test_add():
    fm = FileSet()

    assert len(fm) == 0

    file_a = File(contents='test file a', count=1)
    fm.add('file_a', file_a)

    assert len(fm) == 1

def test_remove():
    fm = FileSet()

    assert len(fm) == 0

    file_a = File(contents='test file a', count=1)
    fm.add('file_a', file_a)

    assert len(fm) == 1

    fm.remove('file_a')

    assert len(fm) == 0

def test_rename():
    fm = FileSet()

    file_a = File(contents='test file a', count=1)
    fm.add('file_a', file_a)

    fm.rename('file_a', 'file_b')

    assert 'file_b' in fm
    assert 'file_a' not in fm

def test_getitem():
    fm = FileSet()
    file_a = File(contents='test file a', count=1)
    fm.add('file_a', file_a)

    fm['file_a'] == file_a


def test_subset_access():
    fm = FileSet()

    fm.add('file_a', File(contents='test file a', count=1))
    fm.add('file_b', File(contents='test file b', count=1))

    fm_sub = fm.filter(['*_a'])

    fm_sub['file_a']['count'] += 1

    assert fm['file_a']['count'] == 2
    assert fm['file_b']['count'] == 1


def test_subset_add():
    fm = FileSet()

    file_b = File(contents='test file b', count=1)
    file_c = File(contents='test file c', count=1)

    fm.add('file_a', File(contents='test file a', count=1))
    fm.add('file_b', file_b)

    fm_sub = fm.filter(['*_a'])

    fm_sub.add('file_c', file_c)

    assert 'file_b' not in fm_sub
    assert 'file_c' in fm_sub
    assert 'file_c' in fm




import codecs
import logging
import os
import shutil

from hana.errors import HanaPluginError

class FileWriter(object):
    def __init__(self, deploy_path, clean=False):
        self._deploy_path = deploy_path
        self.clean = clean
        self.logger = logging.getLogger(self.__module__)

        if not os.path.isdir(self._deploy_path):
            raise DeployDirectoryError()

    def _clean_output_dir(self):
        #TODO: see if we can avoid removing the dir itself
        shutil.rmtree(self._deploy_path)
        self._create_output_dir()

    def _create_output_dir(self):
        os.mkdir(self._deploy_path)

    def __call__(self, files, hana):
        if self.clean and os.path.isdir(self._deploy_path):
            self._clean_output_dir()

        if not os.path.isdir(self._deploy_path):
            self._create_output_dir()

        for filename, f in files:
            output_path = os.path.join(self._deploy_path, filename)

            def makedirs(path, directory):
                if not directory:
                    return

                if os.path.isdir(os.path.join(self._deploy_path, path, directory)):
                    return

                makedirs(*os.path.split(path))

                if os.path.isdir(os.path.join(self._deploy_path, path)) or path == '':
                    dirpath = os.path.join(self._deploy_path, path, directory)
                    os.mkdir(dirpath)
                    return

            makedirs(*os.path.split(os.path.dirname(filename)))

            self.logger.debug('Writing %s (%s)', output_path, 'binary' if f.is_binary else 'text')
            if not f.is_binary:
                codecs.open(output_path, 'w', 'utf-8').write(f['contents'])
            else:
                open(output_path, 'wb').write(f['contents'])



class FileLoaderError(HanaPluginError):
    pass

class DeployDirectoryError(FileLoaderError):
    pass




import logging
import os

import pathspec

from hana.errors import HanaPluginError
from hana.core import FSFile

class FileLoader(object):
    def __init__(self, source_path, ignore_patterns=(), source_file_keyword=None):
        self._source_path = source_path
        self.ignore_patterns = ignore_patterns
        self.source_file_keyword = source_file_keyword
        self.logger = logging.getLogger(self.__module__)

        if not os.path.isdir(self._source_path):
            raise SourceDirectoryError()

    def __call__(self, files, hana):
        ignore_spec = pathspec.PathSpec.from_lines('gitwildmatch', self.ignore_patterns)

        for path, _, sfiles in os.walk(self._source_path):
            for f in sfiles:
                source = os.path.join(path, f)

                if ignore_spec.match_file(f):
                    continue

                filepath = os.path.relpath(source, self._source_path)

                if filepath in files:
                    raise FileExistsError("File {} already exists".format(filepath))

                metadata = {}

                if self.source_file_keyword:
                    metadata[self.source_file_keyword] = source

                files.add(filepath, FSFile(source, **metadata))

class FileLoaderError(HanaPluginError):
    pass

class FileExistsError(FileLoaderError):
    pass

class SourceDirectoryError(FileLoaderError):
    pass




#!/usr/bin/env python

from setuptools import setup
from codecs import open
from os import path

basedir = path.abspath(path.dirname(__file__))

with open(path.join(basedir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='hana',
    version='0.0.dev5',

    description='Static site generator',
    long_description=long_description,

    install_requires=[
        'pathspec>=0.5.2',
        'pyyaml>=3.12',
    ],

    author='Mayo Jordanov',
    author_email='mayo@oyam.ca',

    url='https://github.com/mayo/hana',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Environment :: Console',

        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: System Administrators',

        'License :: OSI Approved :: MIT License',

        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',

        'Programming Language :: Python',

        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
        'Topic :: Software Development',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Code Generators',
    ],

    keywords='hana static site generator processing file',

    packages=[
        'hana',
        'hana.plugins'
    ],

    entry_points={
#        'console_scripts': [
#            'hana = hana.core:main',
#        ],
        'hana.plugins': [
            'file_loader = hana.plugins.file_loader.FileLoader',
            'file_writer = hana.plugins.file_writer.FileWriter',
        ]
    }
)




import pytest
import hana
import hana.plugins.file_loader
import hana.plugins.file_writer

#TODO: test absolute paths
#TODO: test relative paths
def test_validation():
    with pytest.raises(hana.plugins.file_loader.SourceDirectoryError):
        b = hana.Hana()
        b.plugin(hana.plugins.file_loader.FileLoader(source_path='nonexistent_directory'))
        b.build()

    with pytest.raises(hana.plugins.file_writer.DeployDirectoryError):
        b = hana.Hana()
        b.plugin(hana.plugins.file_writer.FileWriter(deploy_path='nonexistent_directory'))
        b.build()

def test_simple():
    b = hana.Hana()

    b.plugin(hana.plugins.file_loader.FileLoader(source_path='test/simple'))
    b.plugin(hana.plugins.file_writer.FileWriter(deploy_path='test/out'))

    b.build()

def test_plugins():
    plugin_executed = {}

    def plugin(files, hana):
        plugin_executed['executed'] = True

    b = hana.Hana()
    b.plugin(plugin)
    b.build()

    assert(plugin_executed)




import codecs
#import importlib
import logging
import os.path
import sys
import yaml

import pathspec
import pkg_resources

from hana import errors

#TODO: this probably shouldn't be here
root_logger = logging.getLogger()
if not root_logger.handlers:
    root_logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    root_logger.addHandler(ch)


class Hana(object):

    def __init__(self, configuration=None):
        self.logger = logging.getLogger(self.__module__)

        # Load configuration first, any parameters override config
        if configuration:
            self._load_configuration(configuration)

        self.plugins = []
        self.metadata = {}

        self.files = FileSet()

    def _load_configuration(self, config_file):
        self.logger.info('Loading configuration from "%s"', config_file)
        config = yaml.safe_load(open(config_file))

        plugins_directory = config.get('plugins_directory')
        if plugins_directory:
            self._setup_plugins(plugins_directory)

        plugins = config.get('plugins')
        if plugins:
            self._load_plugins(plugins)

    def _setup_plugins(self, plugins_directory):
        if not os.path.isdir(plugins_directory):
            raise errors.InvalidPluginsDirectory()

        #TODO: load plugins somehow... how do we discover custom plugins vs installed plugins.
        #      Command plugins vs normal plugins, ... can we do entry points with modules? Not
        #      all plugins from plugins_directory should be loaded, only the required ones

        #try:
        #    fp, filename, description = imp.find_module(self.module_name, [plugins_directory])
        #except ImportError as err:
        #    raise errors.PluginNotFoundError(
        #       'Missing plugins: "{}" not found'.format(missing_plugins)
        #    )

        ## Load the module
        #try:
        #    module_obj = imp.load_module(self.module_name, fp, filename, description)
        #except Exception as err:
        #    self.log.exception('Failed to load module "{module:s}" from file "{path:s}": {err:s}'.format(
        #        module=self.module_name, path=self.module_path, err=err))
        #    raise
        #finally:
        #    fp.close()

        #self.module_bytecode = getattr(module_obj, self.module_name)

        ## Update where the module was loaded from
        #self.module_path = os.path.join(MODULES_PATH, filename)

        #self.module_instance = self.module_bytecode()


    def _load_plugins(self, plugins):
        self.logger.info('Loading plugins')

        # Collect configured plugins
        wanted_plugins = set([p.keys()[0] for p in plugins])

        # Collect all entrypoints for plugins
        available_plugins = {e.name: e for e in pkg_resources.iter_entry_points(group='hana.plugins')}

        # Fail early for missing plugins
        missing_plugins = wanted_plugins - set(available_plugins.keys())
        if missing_plugins:
            raise errors.PluginNotFoundError('Missing plugins: "{}" not found'.format(missing_plugins))

        # Iterate plugins and make new instances
        for plugin_spec in plugins:
            plugin_name, plugin_config = plugin_spec.items()[0]

            # Get the plugin callable
            plugin = available_plugins[plugin_name].load()

            # Initialize the plugin if requested or if parameters passed
            if plugin_config.get('init', False) or 'parameters' in plugin_config:
                plugin = plugin(**plugin_config.get('parameters'))

            self.plugin(plugin, plugin_config.get('patterns'))

    def plugin(self, plugin, pattern=None):
        if isinstance(pattern, str):
            pattern = [pattern]

        self.plugins.append((plugin, pattern))

    def build(self):
        self._process()

    def _process(self):
        for plugin, patterns in self.plugins:
            plugin(self.files.filter(patterns), self)

#TODO: Tileset to make filtering more easier.
#      Filter files based on glob pattern, so in plugins:
#files.filter(glob) returns iterator/generator with matches
#TODO: This is not ideal and error prone, as it sub-calls .add, etc.
#      This should really return a proxy that's smart enough to iterate
#      through matching items, rather than recursively call action methods.
class FileSet(object):

    def __init__(self, parent=None):
        self._files = {}
        self._parent = parent

    def __iter__(self):
        return self._files.iteritems()

    def __len__(self):
        return len(self._files)

    def __contains__(self, filename):
        return filename in self._files

    def __getitem__(self, key):
        return self._files[key]

    def filenames(self):
        """Return all filenames
        """

        return self._files.iterkeys()

    def filter(self, patterns):
        """Return FileSet with subset of files
        """

        if not patterns:
            return self

        fm = FileSet(self)

        match = pathspec.PathSpec.from_lines('gitwildmatch', patterns)

        for filename, f in self:
            if match.match_file(filename):
                fm.add(filename, f)

        return fm

    #TODO: verify base path, resolve relative paths
    def add(self, filename, f):
        self._files[filename] = f

        if self._parent:
            self._parent.add(filename, f)

    #TODO: verify base path, resolve relative paths
    def remove(self, filename):
        self._files.pop(filename)

        if self._parent:
            self._parent.remove(filename)

    def rename(self, filename, new_name):
        self._files[new_name] = self._files.pop(filename)

        if self._parent:
            self._parent.rename(filename, new_name)

#TODO: finish me
class FileSetProxy(object):
    def __init__(self, file_manager, patterns):
        self._file_manager = file_manager
        self._patterns = patterns



class File(dict):

    def __init__(self, *args, **kwargs):
        super(File, self).__init__()
        super(File, self).update(*args, **kwargs)

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, dict(self))

    @property
    def is_binary(self):
        # If we don't have any content, treat it as binary and try to do detection later
        if not self['contents']:
            return True

        return '\0' in self['contents']

    def update(self, *args, **kwargs):
        if args:
            if len(args) > 1:
                raise TypeError("update expected at most 1 arguments, "
                                "got {:d}".format(len(args)))

            other = dict(args[0])

            for key in other:
                self[key] = other[key]

        for key in kwargs:
            self[key] = kwargs[key]

    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        return self[key]


class FSFile(File):

    def __init__(self, filename, *args, **kwargs):
        super(FSFile, self).__init__(*args, **kwargs)

        if not filename:
            raise

        self._is_binary = None
        self.loaded = False
        self.filename = filename

    def __getitem__(self, key):
        if key == 'contents' and not self.loaded:
            self._get_contents()

        return super(FSFile, self).__getitem__(key)

    # Override is_binary to avoid loading files unnecesarily
    @property
    def is_binary(self):
        # If the file is already loaded, it may have been processed
        if self.loaded:
            #TODO: call super.is_binary?
            self._is_binary = None
            return '\0' in self['contents']

        if self._is_binary is None:
            with open(self.filename, 'rb') as fin:
                CHUNKSIZE = 1024

                while True:
                    chunk = fin.read(CHUNKSIZE)
                    if '\0' in chunk:
                        self._is_binary = True
                        break

                    if len(chunk) < CHUNKSIZE:
                        break

                    self._is_binary = False

        return self._is_binary

    def _get_contents(self):
        if self.is_binary:
            self['contents'] = open(self.filename, 'rb').read()

        else:
            self['contents'] = codecs.open(self.filename, 'r', 'utf-8').read()

        self.loaded = True

    def __repr__(self):
        return '<{} {} {}>'.format(self.__class__.__name__, self.filename, dict(self))

